#!/usr/bin/perl -T -w 

# In order for this daemon to work correctly, something like 'mlab-listen 33001/tcp' must be added
#  to /etc/services and the daemon must be enabled in inetd.conf with a line like:
#
# mlab-listen     stream  tcp     nowait/10/3/1   nobody  path-to-mlab-listen-daemon/bin/mlab-listen 

use warnings;
use strict;
use Data::Dumper;
use JSON;
use DBI;
use FindBin qw($Bin);
use Socket;
use Crypt::Rijndael;
use MIME::Base64;

use Log::Log4perl;

my $read_limit = 5000;

my $logger_file = "$Bin/mlab-listen.logger";
if (-f $logger_file) {
        Log::Log4perl->init($logger_file);
} 
else {
        print STDERR "WARNING: Cannot find logger configuration file $logger_file \n";
}
my $logger = Log::Log4perl::get_logger();

#intercept messages going to STDERR 
tie *STDERR, "Trapper";


my $db_driver = "mysql";
my $db_host   = "localhost";
my $db_user   = "mlab-listen";
my $db_passwd = 'add_database_password_here';
my $db_dbase  = "add_database_name_here";
my $db_dsn    = "DBI:$db_driver:database=$db_dbase;host=$db_host";

#connect to the database 
defined ( my $dbh = DBI->connect ($db_dsn, $db_user, $db_passwd) ) or $logger->fatal(DBI::errstr);

#on error, please die 
$dbh->{RaiseError} = 1;

my $reporting_host = 'unknown';
my $reporting_ip;
eval {
        my ($port, $iaddr) = sockaddr_in(getpeername(STDOUT));
        $reporting_ip = inet_ntoa($iaddr);
        $reporting_host = $reporting_ip.':'.$port;
};
if($@) {
        $logger->debug("Connection from a non-socket descriptor");
}
else {
        $logger->debug("Connection from $reporting_host");
}

#confirm that the reporting host actually connects from an ISP contained in our database
if($reporting_host eq 'unknown') {
        $logger->fatal("Unknown reporting host, I cannot register this measurement");
} 
else {
        if( defined($reporting_ip) ) {
                my $sth;
                eval {
                        $sth = $dbh->prepare('SELECT INET_NTOA(ip_start) as start,INET_NTOA(ip_stop) as stop,ts,isp_id FROM ipv4_to_isp WHERE INET_ATON(?) BETWEEN ip_start AND ip_stop');
                        $sth->execute($reporting_ip);
                };
                if($@) {
                        $logger->fatal("Cannot execute IP range query: $@");
                        $dbh->disconnect;
                        exit;
                }

                my $result = $sth->fetchall_hashref(['start','stop']);

                $sth->finish;

                $logger->trace(Dumper($result));
                if( keys %{$result} ) {
                        my @isp_ids;
                        for my $ip_start (keys %{$result}) {
                                for my $ip_stop (keys %{$result->{$ip_start}}) {
                                        push @isp_ids,join(',',(map { $_.'='.$result->{$ip_start}->{$ip_stop}->{$_} } sort keys %{$result->{$ip_start}->{$ip_stop}} ));
                                }
                        }
                        $logger->debug("Matching IP blocks: ".join('  ',@isp_ids)); 
                }
                else {
                        $logger->fatal("User IP $reporting_ip not found in our database...rejecting the measurement");
                        $dbh->disconnect;
                        exit;
                }

        } 
        else {
                $logger->fatal("Error! reporting_ip seems to be undefined...looks like an internal error.");
                exit;
        }
}




if(!defined(syswrite(STDOUT,"ok\n"))) {
        $logger->trace("$reporting_host: Cannot tell ok to $reporting_host");
}
close(STDOUT);

#my $slurp = do { local $/; <> };
my $overall_read_size = 0;
my $slurp;
while (1) {
        my $read_size = sysread STDIN,$slurp,1024,$overall_read_size;
        if (!defined($read_size)) { 
                $logger->fatal("Connection from $reporting_host broken after reading $overall_read_size bytes. Error was: $!"); 
                exit;
        }
        last if ($read_size == 0);
        $overall_read_size += $read_size;
        $logger->trace("$reporting_host: read $read_size bytes, overall is $overall_read_size");
        if($overall_read_size > $read_limit) {
                $logger->fatal("read limit $read_limit exceeded. Closing connection with peer $reporting_host");
                exit;
        }
} 

my $data;

$logger->trace("Input data is:\n------------\n$slurp\n------------");

eval {
        $data = from_json($slurp);
};
if($@) {
        $logger->fatal($@);
        exit;
}

#a little trick, can be also fixed applet-side. 
my %data2 = map { my $v = $data->{$_}; $_ =~ s/:$//g; $_=>$v  } (keys %{$data});
$data = \%data2;


my $record_str = join(',',(map {$_.'='.$data->{$_}} keys %{$data}));
$logger->debug("record is ",Dumper($data));

if(! defined($data->{measurement_tool}) ) {
        $logger->fatal("measurement tool not defined");
        exit;
} 
elsif($data->{measurement_tool} eq 'ndt') {
        $logger->debug("Measurement tool is NDT");
        $logger->debug("Inserting record: $record_str");
        my $user_id = decode_id($data->{user_id});
        $logger->debug("user id is ",$user_id);
        eval {
                my @args = (
                                        $reporting_host,
                                        $user_id,
                                        $data->{connection_id},
                                        $data->{measurement_tool},
                                        $data->{version},
                                        $data->{report_host},
                                        $data->{report_port},
                                        $data->{upstream_bw},
                                        $data->{downstream_bw},
                                        $data->{rtt},
                                        $data->{loss},
                                        $data->{jitter},
                );
                #$logger->debug(Dumper(\@args));
                ###$dbh->trace(2);
                my $sth = $dbh->prepare('insert into generic_measurement set 
                                        reporting_host=?, 
                                        user_id=?, 
                                        connection_id=?,
                                        measurement_tool=?,
                                        version=?,
                                        report_host=?,
                                        report_port=?,
                                        upstream_bw=?,
                                        downstream_bw=?,
                                        rtt=?,
                                        loss=?,
                                        jitter=?
                                        '
                                        );
                $sth->execute(@args);



                ####my $str = qq(insert into generic_measurement set reporting_host='$reporting_host',user_id='$user_id',connection_id='$data->{connection_id}',measurement_tool='$data->{measurement_tool}',version='$data->{version}',report_host='$data->{report_host}',report_port='$data->{report_port}',upstream_bw='$data->{upstream_bw}',downstream_bw='$data->{downstream_bw}',rtt='$data->{rtt}',loss='$data->{loss}',jitter='$data->{jitter}');
                ####$logger->debug($str);
                ####my $sth = $dbh->prepare($str);
                ####$sth->execute;

                my $insertid = $dbh->{mysql_insertid};
                $logger->debug("Successfully inserted row number $insertid into generic_measurement");


                my @args2 = (
                        $insertid,
                        $reporting_host,
                        $user_id,
                        $data->{connection_id},
                        $data->{measurement_tool},
                        $data->{version},
                        $data->{report_host},
                        $data->{report_port},
                        $data->{os_name},
                        $data->{os_version},
                        $data->{os_architecture},
                        $data->{java_version},
                        $data->{java_vendor},
                        $data->{is_application},
                        $data->{upstream_bw},
                        $data->{downstream_bw},
                        $data->{rtt},
                        $data->{loss},
                        $data->{jitter},
                        $data->{mss},
                        $data->{out_of_order},
                        $data->{sack_blocks},
                        $data->{sack_enabled},
                        $data->{nagle_enabled},
                        $data->{ECN_enabled},
                        $data->{time_stamping_enabled},
                        $data->{timeouts},
                        $data->{retransmissions},
                        $data->{duplicate_acks},
                        $data->{ccip},
                        $data->{scip},
                        $data->{ssip},
                );
                $sth = $dbh->prepare('insert into ndt_measurement set 
                                        measurement_id=?,
                                        reporting_host=?, 
                                        user_id=?, 
                                        connection_id=?,
                                        measurement_tool=?,
                                        version=?,
                                        report_host=?,
                                        report_port=?,
                                        os_name=?,
                                        os_version=?,
                                        os_architecture=?,
                                        java_version=?,
                                        java_vendor=?,
                                        is_application=?,
                                        upstream_bw=?,
                                        downstream_bw=?,
                                        rtt=?,
                                        loss=?,
                                        jitter=?,
                                        mss=?,
                                        out_of_order=?,
                                        sack_blocks=?,
                                        sack_enabled=?,
                                        nagle_enabled=?,
                                        ECN_enabled=?,
                                        time_stamping_enabled=?,
                                        timeouts=?,
                                        retransmissions=?,
                                        duplicate_acks=?,
                                        ccip=?,
                                        scip=?,
                                        ssip=?
                                        '
                                        );
                $sth->execute(@args2);
                $logger->debug("Successfully inserted row number $insertid into ndt_measurement");

                #and now, the final chapter. Insert all web100 variables available
                $logger->debug('inserting everything to web100_measurement');
                my @fields = qw(measurement_id reporting_host user_id connection_id measurement_tool version report_host report_port os_name os_version os_architecture java_version java_vendor is_application upstream_bw downstream_bw rtt loss jitter mss out_of_order sack_blocks sack_enabled nagle_enabled ECN_enabled time_stamping_enabled timeouts retransmissions duplicate_acks ccip scip ssip);
                my @web_100_fields = grep { /^web100_/ } (keys %{$data});
                push @fields,( @web_100_fields );
                $logger->debug('fields are '.join(',',@fields));
                push @args2,(map { $data->{$_} } @web_100_fields);
                $logger->debug('values are '.join(',',@args2));
                my $to_prepare = 'INSERT INTO web100_measurement set '.join(',',(map {"`$_`=?"} @fields));
                #$logger->debug($to_prepare);
                $sth = $dbh->prepare($to_prepare);
                $sth->execute(@args2);


        };#end of eval
        if($@) {
                $dbh->disconnect;
                $logger->fatal("Cannot insert record into database: $@");
                exit;
        }
}
else {
        $logger->fatal("Cannot (yet) handle messages from  ",$data->{measurement_tool});
        exit;
}

$dbh->disconnect;
$logger->debug("Finished OK");


sub decode_id {
        defined(my $message = shift(@_) ) or do { $logger->fatal('incorrect call to decode_id without argument');exit};

        $message = decode_base64($message);

        my $encoded = substr $message,0,48;
        my $iv = substr $message,48,16;

        my $cipher = Crypt::Rijndael->new( '**************************' , Crypt::Rijndael::MODE_CBC() );
        $cipher->set_iv($iv);

        my $plaintext = $cipher->decrypt($encoded);

        $logger->trace("plaintext is [$plaintext]");
        $logger->trace("length is ".length($plaintext));

        #$logger->debug(ord(substr($plaintext,length($plaintext),1)));

        if(length($plaintext) != 48) {
                $logger->fatal('Error! Plaintext is not 48 bytes long!');
                exit;
        }


        $logger->debug("plaintext is [$plaintext]");
        my ($id) = ($plaintext =~ /**********\s+(\d+)***********/); 
        if (! defined($id) ) {
                $logger->fatal("Error! Cannot decode string from [$plaintext]. ");
                exit;
        }
        #$logger->debug("returning id = [$id]");
        return $id;
}



#copied from Log4Perl faq
########################################
package Trapper;
########################################

###use Log::Log4perl qw(:easy);

sub TIEHANDLE {
        my $class = shift;
        bless [], $class;
}

sub PRINT {
        my $self = shift;
        $Log::Log4perl::caller_depth++;
        $logger->debug(@_);
        $Log::Log4perl::caller_depth--;
}

1;